from astropy.io import fits
import astropy.units as u
from numpy import ma
import numpy as np
import scipy as sp
from lmfit import models
import matplotlib.pyplot as plt

from .data import DataND
from .world import Position, VelWave
from .spatial import SpatLine
from .spectral import SpecLine
from .plotter import (ImCoords, get_plot_norm, get_plot_extent,
                      get_background_rms, get_contour_levels,
                      configure_axes)
from .lines import lines
from .tools import parse_badlines

from datetime import datetime
from tqdm import trange

class PVSlice(DataND):

    '''
    This is to just manage the data shit, but it might
    end up being completely superfluous. We'll find out.
    '''

    def __getitem__(self, item):
        """
        Return an object:
        pvslice[i, j] = value
        pvslice[:, j] = spatial profile
        pvslice[i, :] = spectral profile
        pvslice[:, :] = sub-pvslice
        """
        obj = super(PVSlice, self).__getitem__(item)
        if isinstance(obj, DataND):
            if obj.ndim == 2:
                return obj
            elif obj.ndim == 1 and obj._is_spatial:
                cls = SpatLine
            elif obj.ndim == 1 and obj._is_spectral:
                cls = SpecLine
            return cls.new_object(obj)
        else:
            return obj

    def spectral_window(self, vmin, vmax=None, unit=None):
        '''
        Get a small view of the velocity/wavelength range.

        Parameters
        -----------
        vmin : float
            lower bound; if vmax is None, only a single pixel will be returned
        vmax : float or None
            upper bound of the view range
        unit : `astropy.units.Unit` or None
            if unit is None, vmin and vmax will be treated as pixels! otherwise,
            give it a velocity or wavelength value

        Returns
        --------
        out : `telassar.PVSlice` object
        '''

        if self.velwave.unit is None:
            raise ValueError("We need coordinates along the spectral direction")

        if vmax is None:
            vmax = vmin

        if unit is None:
            pmin = max(0, int(vmin + 0.5))
            pmax = min(self.shape[1], int(vmax + 0.5))

        else:
            pmin = max(0, self.velwave.wav2pix(vmin, nearest=True))
            pmax = min(self.shape[1], self.velwave.wav2pix(vmax, nearest=True) + 1)

        return self[:, pmin:pmax]

    def spatial_window(self, amin, amax=None, unit=None):
        '''
        Return a view of a spatial window

        Parameters
        ----------
        amin : float
            the lower bound of the view; assumed in arcseconds
        amax : float or None
            the upper bound of the view. if None, only a single pixel value will
            be returned (in arcseconds). NOTE: this pixel value will span the
            entire wavelength range!
        unit : `astropy.units.Unit` or None
            if no unit is specified, values will be treated like pixels. otherwise,
            give it a u.arcsec value or something.

        Returns
        ---------
        out : `telassar.PVSlice` obj
        '''

        if self.position.unit is None:
            raise ValueError("We need coordinates along the spatial direction")

        if amax is None:
            amax = amin

        if unit is None:
            pmin = max(0, int(amin + 0.5))
            pmax = min(self.shape[0], int(amax + 0.5))

        else:
            pmin = max(0, self.position.offset2pix(amin, nearest=True))
            pmax = min(self.shape[0], self.position.offset2pix(amax, nearest=True) + 1)

        return self[pmin:pmax, :]

    def spatial_profile(self, arc, wave, spat_unit=False, spec_unit=False):

        """
        Extract a 1D spatial profile from a position-velocity slice.

        Parameters
        -----------
        arc : array-like
            the region in distance from which to extract a profile
        wave : array-like
            the wavelength/velocity range over which to sum the profile
        spat_unit : bool
            toggles whether to use pixels or the native unit for distance;
            default is False (uses pixel)
        spec_unit : bool
            toggles whether to use pixels or native unit for wavelength/velocity;
            default is False (uses pixel)

        Returns
        -----------
        out : `telassar.SpatLine` object
        """

        if (isinstance(arc, int) or isinstance(wave, int) or len(arc) != 2 or
                len(wave) !=2):
            raise ValueError("Can't extract profile with only one point!")

        # get the spatial and spectral limits in pixel or arcseconds
        if spat_unit:
            pmin = max(0, self.position.offset2pix(arc[0], nearest=True))
            pmax = min(self.shape[0], self.position.offset2pix(arc[1], nearest=True))
        else:
            pmin = max(0, int(arc[0] + 0.5))
            pmax = min(self.shape[0], int(arc[1] + 0.5))

        if spec_unit:
            lmin = max(0, self.velwave.wav2pix(wave[0], nearest=True))
            lmax = min(self.shape[1], self.velwave.wav2pix(wave[1], nearest=True))
        else:
            lmin = max(0, int(wave[0] + 0.5))
            lmax = min(self.shape[1], int(wave[1] + 0.5))


        sx = slice(pmin, pmax+1)
        sy = slice(lmin, lmax+1)

        res = self.data[sx, sy].sum(axis=1)
        wcs = self.position[sx]
        return SpatLine(data=res, wcs=wcs, unit=self.position.unit)

    def spectral_profile(self, wave, arc, spec_unit=False, spat_unit=False):

        """
        Extract a 1D spectral profile from a position-velocity slice

        Parameters
        ----------
        wave : array-like
            the spectral region over which to extract the profile
        arc : array-like
            the distance/offset range over which to sum the profile
        spec_unit : bool
            toggles whether to use pixels or native units for the spectral region;
            default is False (use pixels)
        spat_unit : bool
            toggles pixels/native units for distance region; default is False

        Returns
        ----------
        out : `telassar.SpecLine` object
        """

        if (isinstance(wave, int) or isinstance(arc, int) or len(wave) != 2 or
                len(arc) !=2):
            raise ValueError("Can't extract profile with only one point!")

        # get the spectral and spatial limits in pixel or native unts
        if spec_unit:
            lmin = max(0, self.velwave.wav2pix(wave[0], nearest=True))
            lmax = min(self.shape[1], self.velwave.wav2pix(wave[1], nearest=True))
        else:
            lmin = max(0, int(wave[0] + 0.5))
            lmax = min(self.shape[1], int(wave[1] + 0.5))

        if spat_unit:
            pmin = max(0, self.position.offset2pix(arc[0], nearest=True))
            pmax = min(self.shape[0], self.position.offset2pix(arc[1], nearest=True))
        else:
            pmin = max(0, int(arc[0] + 0.5))
            pmax = min(self.shape[0], int(arc[1] + 0.5))

        sx = slice(pmin, pmax+1)
        sy = slice(lmin, lmax+1)

        res = self.data[sx, sy].sum(axis=0)
        spec = self.velwave[sy]

        return SpecLine(data=res, spec=spec, unit=self.velwave.unit)

    def plot(self, scale='linear', ax=None, fig_kws=None, imshow_kws=None,
             vmin=None, vmax=None, zscale=None, emline=None):
        '''
        This function generates an simple plot of the desired data.

        Parameters
        ----------
        scale : str
            the interpolation style desired. default is linear, but can accept
            others from the list:
                ['linear', 'log', 'asinh', 'arcsinh', 'sqrt']
        ax : None or `matplotlib.pyplot.axes` instance
            instanciate plotter with an axis?
        ax_kws : None or dict
            keywords to be passed to the `plt.subplots()` routine
        imshow_kws : None or dict
            keywords to be passed to the `plt.imshow()` routine
        vmin : None or float
            minimum value for plotting normalization
        vmax : None or float
            maximum value for plotting normalization
        zscale : None or str
            do you want a zscale normalization?
        emline : None or str
            can optionally pass an emission line name, and the list of lines
            from above is checked to create a pretty title

        '''

        if fig_kws is None:
            fig_kws = {'figsize' : (6, 9)}
        # TODO: set some defaults here
        if imshow_kws is None:
            imshow_kws = {}

        if emline is not None:
            if emline in lines.keys():
                emis = lines[emline][2]

        # set the data and plot parameters
        res = self.copy()
        data = self.data.copy()

        if ax is None:
            fig, ax = plt.subplots(**fig_kws)

        # get a norm
        norm = get_plot_norm(data, vmin=vmin, vmax=vmax, zscale=zscale,
                             scale=scale)
        # set the extent of the data
        extent = get_plot_extent(self.position, self.velwave)

        cax = ax.imshow(data, interpolation='nearest', origin='lower',
                        norm=norm, extent=extent, **imshow_kws)

        # format the axes and coordinates
        toggle_unit = True if extent is not None else False
        configure_axes(ax, self)
        ax.format_coord = ImCoords(res, data, toggle_unit)
        fig.subplots_adjust(left=0.15, right=0.85)

        return cax

    def plot_contours(self, figure=None, place=None, sigma=None, mask=None,
                      levels1=None, levels2=None, cmaps=None, fig_kws=None,
                      plt_kws=None, emline=None):
        '''
        Generate a contour plot of the data. Useful for jet visualization!

        Parameters
        -----------
        figure : None or `matplotlib.Figure` instance
            if you have a figure instance you want to send this to, specify it
        sigma : None or float
            the basis for generating levels. a 3sigma value indicates detection
            of a source, we abbreviate it here to just sigma
        place : int
            if you have multiple subplot axes, you can send the image to one
        mask : `np.ma.masked_array`
            if you want to specify a mask to send when computing the background
            levels, do it here
        levels1 : None or `np.ndarray` or list
            the contour levels for the jets
        levels2 : None or `np.ndarray` or list
            the contour levels for the background
        cmaps : None or `matplotlib.colors.Colormap`
            the list of colormaps to pass to `plt.contour`
        emline : str
            if your emission line of interest is in `lines.py` this will do
            a pretty formatting and set it as the figure title

        The `fig_kws` and `plt_kws` are just keyword dictionaries to pass to
        the plotter
        '''

        if fig_kws is None:
            fig_kws = {'figsize' : (5, 9)}

        emis = None
        if emline is not None:
            if emline in lines.keys():
                emis = lines[emline][2]
            else:
                emis = emline

        # default cmap colors
        colors = ['gist_gray', 'Oranges', 'gray']
        data = self.data.copy()
        # generate a sigma based on the data?
        if sigma is None:
            sigma = 3.
        sig = get_background_rms(data, sigma=sigma, N=10, mask=None)

        if (levels1 is None) or (levels2 is None):
            lvls1, lvls2 = get_contour_levels(data, sig)

        levels1 = levels1 if levels1 is not None else lvls1
        levels2 = levels2 if levels2 is not None else lvls2

        cmaps = colors if cmaps is None else cmaps

        # get an extent
        if plt_kws is None:    
            ext = get_plot_extent(self.position, self.velwave)
            plt_kws = {'extent': ext}

        # make the plot
        if figure is not None:
            fig = plt.gcf()
            if place < len(fig.axes):
                ax = fig.axes[place]
        else:
            fig, ax = plt.subplots(**fig_kws)
        if emis is not None:
            ax.set_title(rf'{emis}', fontsize=14)

        jet1 = ax.contour(data, levels=levels1, cmap=cmaps[0], **plt_kws)
        jet2 = ax.contourf(data, levels=levels1, cmap=cmaps[1], **plt_kws)
        bkgrd = ax.contourf(data, levels=levels2, cmap=cmaps[2], **plt_kws,
                            alpha = 0.8)

        # format the canvas and coordinates
        configure_axes(ax, self)
        toggle_unit = True
        ax.format_coord = ImCoords(self, data, toggle_unit)

        return ax

    def moments(self, units=False):
        '''
        Return [y_width, x_width] moments (order=1) of a 2D gaussian
        Essentially the same as the example from the SciPy Cookbook:
        https://scipy-cookbook.readthedocs.io/items/FittingData.html

        Parameters
        ----------
        units : bool
            if True, convert the widths to units; otherwise treat them as pixels

        Returns
        ----------
        out : `np.ndarray`
        '''

        # use absolute values to ensure no issues with sqrt
        total = np.abs(data).sum()
        Y, X = np.indices(data.shape)
        y = np.argmax((X * np.abs(data)).sum(axis = 1) / total)
        x = np.argmax((Y * np.abs(data)).sum(axis = 0) /total)
        col = data[int(y), :]
        row = data[:, int(x)]
        xwidth = np.sqrt(np.abs((np.arange(col.size) - y)*col).sum() /
                         np.abs(col).sum())
        ywidth = np.sqrt(np.abs((np.arange(row.size) - x)*row).sum() /
                         np.abs(row).sum())
        height = data.max()
        #mom = np.array([ywidth, xwidth])
        return height, y, x, ywidth, xwidth

    def update_header(self):
        '''Format a new header'''
        hdr = self.header.copy()
        hdr['date'] = (str(datetime.now()), 'Creation Date')
        hdr['author'] = ('TELASSAR', 'Origin of the file')

        # if the data array has been altered, update the header
        noff, nlbda = self.shape
        crval1 = self.velwave.get_start()
        crpix1 = 1.
        # let's keep the starting pixel where the world coord is 0
        crpix2 = round(self.position.offset2pix(0), 2)
        crval2 = 0.
        cdelt1 = round(self.velwave.get_step(), 2)
        cdelt2 = round(self.position.get_step(), 2)

        hdr['NAXIS1'] = (nlbda, "Length of data axis 1")
        hdr['NAXIS2'] = (noff, "Length of data axis 2")
        hdr['CRPIX1'] = (crpix1, "Pixel coordinate at reference point")
        hdr['CRPIX2'] = (crpix2, "Pixel coordinate at reference point")
        hdr['CRVAL1'] = (crval1, "Coordinate value at reference point")
        hdr['CRVAL1'] = (crval2, "Coordinate value at reference point")
        hdr['CDELT1'] = (cdelt1, "Coordinate increment at reference point")
        hdr['CDELT2'] = (cdelt2, "Coordinate increment at reference point")
        hdr.pop('CUNIT1')

        return hdr

    def to_fits(self, fname):

        new_hdr = self.update_header()

        hdul = fits.PrimaryHDU(data=self.data.data, header=new_hdr)
        hdul.write("fname.fits")

    def radial_velocity(self, ref, lbdas, vcorr=None, unit='angstrom',
                        nearest=False):
        '''
        Compute the radial velocity of an emission range based on some
        reference emission and wavelength array.

        Parameters
        -----------
        ref : float or str
            the emission line in vaccum or air. if float, just use the number; if
            str, then it needs to be something from the `lines` list at the top
        lbdas : list, tuple, or `np.ndarray`
            the wavelength range for which to compute the velocity.
        vcorr : float
            the velocity correction; None by default
        Returns
        -----------
        velocities : `np.ndarray`
            array of velocities (blue- and red-shifted) computed from arguments
        '''

        from astropy.constants import c

        if unit.lower() == 'angstrom':
            if nearest:
                px = self.velwave.wav2pix(lbdas, nearest=True)
                obs = self.velwave.pix2wav(px)
            else:
                obs = np.atleast_1d(lbdas)
        elif unit.lower() == 'pixel':
            obs = self.velwave.pix2wav(lbdas)
        else:
            raise Exception("unit must be pixel or angstrom")

        # Get the wavelength array

        # if vcorr is supplied, give it units; else set to 0
        if vcorr is None:
            vcorr = 0
        vcorr *= u.km / u.s

        # is the ref line in the list?
        try:
            for line, val in lines.items():
                if ref == line:
                    ref = val[0]

            # get the doppler shifted velocity
            vrad = c.to('km/s') * (obs - ref) / ref


            # If vcorr is supplied, note that it is multiplicative, not additive
            # so it should be expressed as:
            # v_t = v_m + v_b + (v_b * v_m) / c
            # see Wright & Eastman (2014)
            # https://ui.adsabs.harvard.edu/abs/2014PASP..126..838W/abstract
            vtrue = vrad + vcorr + (vcorr * vrad) / c

        except TypeError:
            return 'Line Not Found!'

        # # estimate the uncertainties
        # #   dv = c * dlbda/lbda
        # # where lbda is the rest wavelength of the reference
        # # and dlbda is HWHM
        # step = self.velwave.get_step()
        # dv = c.to('km/s') * (step / ref)

        return vtrue

    def get_flux(self, arc, wave):

        lpix = self.velwave.wav2pix(wave, nearest=True)
        apix = self.position.offset2pix(arc, nearest=True)

        return self.data[apix, lpix]

    def register_skylines(self, badlines=None):
        '''
        Emulating IRAF a bit here: register a skylines file to mask
        specified regions. if no file is specified, this will look
        in the working direcetory and try to find one with the name
        `badlines.dat`
        '''
        import pathlib

        # does a skylines attribute already exist?
        skylines = self.skylines if self.skylines else {}

        # look only in the current working directory
        workdir = pathlib.Path.cwd()

        # if no data file is given, look for it
        if badlines is None:
            look_for = "badlines.dat"
            path = workdir / look_for

            if path.exists():
                for emis, l1, l2 in parse_badlines(path.name):
                    skylines[emis] = [l1, l2]
            else:
                self._logger.warning("No badlines.dat file found!")
                self._logger.warning("Unable to register skylines.")
        else:
            # Does it look like a UNIX path?
            if badlines.startswith("~"):
                badlines = pathlib.Path(badlines).expanduser()

            # look for it in the working directory first
            if (workdir / badlines).exists():
                path = workdir / badlines
            elif badlines.exists():
                path = badlines
            else:
                self._logger.warning("No badlines.dat file found!")
                self._logger.warning("Unable to register skylines.")
            if path:
                for emis, l1, l2 in parse_badlines(path):
                    skylines[emis] = [l1, l2]

        self.skylines = skylines if skylines else None

    def unregister_skylines(self, key=None):
        '''Remove skyline info from the instance'''

        # if a key is passed, is it in the sky dict?
        if key is not None:
            try:
                self.skylines.pop(key)
                self._logger.info("Unregistering skyline %s" % key)
            except Exception:
                self._logger.warning("%s not in registry!" % key)
                self._logger.warning("No skyline info unregistered",
                                    exc_info=True)
        else:
            self._logger.info("Unregistering all skyline info!")
            self.skylines = None

    def _interp_skylines(self, pix):
        '''
        Utilize `scipy.interpolate.splrep` and `scipy.interpolate.splev`
        to interpolate the masked skyline values

        Parameters:
        ----------
        pix : `int`
            The pixel index at which to perform the interpolation
        '''
     
        # get the spectrum and wavelength array at the given pixel
        raw = self.data[pix, :].copy()
        lbda = self.velwave.pix2wav()

        # pad the compressed array
        line = np.pad(raw.compressed(), 1, 'edge')

        # what are the masked wavelengths?
        masked = lbda[raw.mask]

        # get the effective wavelength range of the data
        w1, w2 = self.velwave.pix2wav([-0.5, self.shape[1] - 0.5])

        # concatenate the effective and compressed wavelengths
        wave = np.concatenate(([w1], np.compress(~raw.mask, lbda), [w2]))

        # get the spline representation
        tck = sp.interpolate.splrep(wave, line)

        # evaluate it
        filled = sp.interpolate.splev(masked, tck)

        return filled

    def skymask(self, wave=None, spec_unit=u.angstrom, verbose=False):
        '''
        Mask skyline emission over the given range. Note that this
        assumes the skyline covers the entire spatial range, so use
        it with caution. If reading from the `skylines` attribute,
        it assumes the values are stored in wavelengths, not pixels.

        TODO: Add support for specifying spatial extents, and add an
        option for the `skylines` dict to toggle pixels/wavelength.

        Parameters:
        -----------
        wave : list or tuple, optional
            The wavelength range to mask; if None, try to read from
            the `self.skylines` dict
        spec_unit : astropy.unit.Unit, optional
            self-explanatory; if None, any given `wave` value is treated
            as pixels
        verbose : bool
            if True, print a logger line for each registered skyline masked
        '''

        # first see if any skylines are registered
        if self.skylines:
            is_registered = True
        else:
            is_registered = False

        if not is_registered:
            if wave is None:
                self._logger.warning("No skylines are registered and no wavelength"
                                     "is specified. Aborting procedure.")
                return
            else:
                wave = np.asarray(wave)

                if len(wave) != 2:
                    self.logger.warning("Only two wavelength values can be "
                                        "specified at a time!")
                    return

                if spec_unit:
                    p1, p2 = self.velwave.wav2pix(wave, nearest=True)
                else:
                    p1, p2 = wave.astype(int)

            # mask it
            self.data[:, p1:p2] = ma.masked

        elif is_registered:
            # get the wavelengths from the skylines dict
            for k, *v in self.skylines.items():
                if verbose:
                    self.logger.info("Masking %s line..."%k)

                p1, p2 = self.velwave.wav2pix(*v, nearest=True)
                self.data[:, p1:p2] = ma.masked

    def skysub(self, arcs=None, unit=u.arcsec, verbose=False, progress=False):
        '''
        Iterate over spatial pixels and perform a spline interpolation
        of the masked skyline regions. This calls `_interp_skylines()`.

        Parameters:
        -----------
        arcs : list or tuple, optional
            pixel or offset values over which to iterate; if None,
            operate on the entire spatial range
        unit : `astropy.units.Unit`, optional
            if None, assume `arcs` is in pixels
        verbose : bool
            if True, print a logger line for each sky emission 
        progress : bool
            if True, use `tqdm` to print a progress bar
        '''

        if arcs is not None:
            arcs = np.asarray(arcs)
            if unit:
                arcs = self.position.offset2pix(arcs, nearest=True)
            else:
                arc = arcs.asype(int)
        else:
            arcs = np.array([0, self.shape[0]])

        p1, p2 = arcs
        
        if progress:
            self._logger.info("Interpolating masked regions...")
            f = trange
        else:
            f = np.arange
        
        for p in f(p1, p2):
            # TODO: find a better way to do this
            data = self.data[p, :]
            data[data.mask] = self._interp_skylines(pix=p)
            self.data[p, :] = data



from astropy.io import fits
import astropy.units as u
from numpy import ma
import numpy as np
from lmfit import models
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator

from .data import DataND
from .world import Position, VelWave
from .plotter import *
from .tools import is_notebook
from .lines import lines
from .fitter import Modeller

class SpatLine(DataND):

    _is_spatial = True

    def plot(self, **kwargs):
        """
        Placeholder for the moment: sort out plot interactivity etc

        NOTE: calling this function will clear the `_coords` attribute, so be
        careful there.
        """
        if is_notebook():
            self._logger.info('Click on data points to save coordinates.')
        else:
            self._logger.info("To enable point selection on the plot, press "
                          #"'ctrl+p'. Press 'ctrl+q' to disable.")
                              "'ctrl+p'. Press 'ctrl+q' to disable.")

        unit_fmt = u.Unit(self.position.unit).to_string('latex')
        xlab = f'Offset ({unit_fmt})'

        coords = []
        # Set up some basic plot interactivity / messages and allow
        # the user to press a key to enter clickable plot mode and
        # save the clicked coordinates to a list
        def on_move(event):
            if event.inaxes is not None:
                xc, yc = event.xdata, event.ydata
                try:
                    i = self.position.offset2pix(xc, nearest = True)
                    x = self.position.pix2offset(i)
                    event.canvas.toolbar.set_message(
                        'xc=%g yc=%g i=%d dist=%g data=%g' %
                        (xc, yc, i, x, self._data[i]))
                except Exception as e:
                    print(e)
                    pass

        def on_click(event):
            
            xc, yc = event.xdata, event.ydata
            
            try:
                i = self.position.offset2pix(xc, nearest = True)
                x = self.position.pix2offset(i)
                data = self._data[i]
                print(f'arc={x}, flux={data}')
                coords.append((x, data))
            
            except Exception as e:
                print(e)
                pass

        def on_key(event):
            
            if event.key == 'a':
                self._logger.info('Enabling point-selection mode...')
                cid = fig.canvas.mpl_connect('button_press_event', on_click)
            if event.key == 'q': 
                self._logger.info("Point-selection mode disabled.")
                cid = fig.canvas.mpl_connect('button_press_event', on_click)
                fig.canvas.mpl_disconnect(cid)

        fig, ax = plt.subplots(figsize=(9, 5))
        xarr = self.position.pix2offset()
        data = self.data.copy()
        kwargs.update({'drawstyle' : 'steps-mid', 'linewidth': 1})

        ax.plot(xarr, data, **kwargs)
        ax.set_xlabel(rf'{xlab}')
        ax.set_ylabel(r'Flux ($F_{\lambda}$)')
        plt.connect('motion_notify_event', on_move)
        if is_notebook():
            cid = fig.canvas.mpl_connect('button_press_event', on_click)
        else:
            cid = fig.canvas.mpl_connect('key_press_event', on_key)

        self._coords = coords

    def integrate(self, amin=None, amax=None, unit=None):
        """
        This will integrate the (non-corrected) flux over a spectral range using
        a simple Simpson's Rule.

        I need to work out the units, but for now this will just return a
        result derived from the native units of the data. Note however that
        the flux over a range will be something like erg/cm^2/s and
        there are no corrections for ie reddening, extinction, etc.

        As of now, the data contains no flux units so the result will be
        simply `u.dimensionless_unscaled * u.unit`. Alternatively, if the units
        of flux are known, simply send them as a keyword when instanciating the
        object or assign the attribute manually, ie `obj.flux_unit = unit`

        Parameters
        ----------
        lmin : float
            lower bound of the velocity or wavelength; if None, the function
            begins at first pixel
        lmax : float
            upper bound; if None, this is the last pixel
        unit : `astropy.units.Unit` or None
            the spectral units of [lmin, lmax]; None by default if these are
            pixel indices

        Returns
        -----------
        out : we'll see
        """

        # get the indices of the specified ranges and all that
        # i1 is index; l1 is wave/velocity
        if amin is None:
            i1 = 0
            amin = self.position.pix2offset(-0.5)
        else:
            if unit is None:
                i1 = amin
                amin = self.position.offset2pix(max(-0.5, i1))
            else:
                i1 = self.position.offset2pix(amin, nearest=False)
            i1 = max(0, int(i1))

        if amax is None:
            i2 = self.shape[0]
            amax = self.position.pix2offset(i2 - 0.5)
        else:
            if unit is None:
                i2 = lmax
                lmax = self.position.pix2offset(min(self.shape[0] - 0.5, i2))
            else:
                i2 = self.position.offset2pix(amax, nearest=False)
            i2 = min(self.shape[0], int(i2) + 1)

        # to work around the array limits, we'll take the lower wavelength or
        # velocity of each pixel + 1 pixel at the end
        d = self.position.pix2offset(-0.5 + np.arange(i1, i2+1))

        # truncate or extend the first and last pixels to the start/end of
        # the values in the spectrum
        d[0] = amin
        d[-1] = amax

        if unit is None:
            unit = self.position.unit

        # get the data over the range
        data = self.data[i1:i2]

        # do the units agree?
        if unit in self.flux_unit.bases:
            out_unit = self.flux_unit * unit
        else:
            try:
                # sort out flux density
                wunit = (set(self.flux_unit.bases) &
                         set([u.pm, u.angstrom, u.nm, u.um])).pop()

                # scale the wavelength axis
                d *= unit.to(wunit)

                # final units
                out_unit = self.flux_unit * wunit

            # if there's an error anywhere, just return unchanged units
            except Exception:
                out_unit = self.flux_unit * unit

        # standard integration: each pixel value is multiplied by the difference
        # in wavelength from the start of the pixel to the start of the next
        flux = (data * np.diff(d)).sum() #* out_unit

        return flux

    def test_fitter(self, model_list, coords = None, plot = True):

        print("We're running a test to send to `fitter.py`")

        my_model = Modeller(self)
        result = my_model.fit_model(model_list, coords=coords, mode='components',
                      plot=plot, densify=10, emline=None, fig_kws=None, ax_kws=None)

        print(my_model.info())



from astropy.io import fits
import astropy.units as u
from numpy import ma
import numpy as np
from lmfit import models
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator

from .data import DataND
from .world import Position, VelWave
from .plotter import *
from .tools import is_notebook
from .lines import lines

import logging

class Modeller:

    def __init__(self, object, **kwargs):

        self._logger = logging.getLogger(__name__)

        self.data = object.data
        self.mask = object.mask

        if object._is_spectral:
            wcs = object.velwave

        if object._is_spatial:
            wcs = object.position

        self.wcs = wcs
        self.unit = wcs.unit

        # convert the flux?
        self.flux = u.Unit('erg/cm**2/s/Angstrom')#object.flux_unit

        if hasattr(object, "_coords") and object._coords is not None:
            self._coords = object._coords
        else:
            self._coords = None

        self.shape = object.shape

    def wcs2pix(self, val, nearest = False):
        """
        quick methods to prevent a lot of if/else try/except statements
        """
        x = np.atleast_1d(val)

        # tell world2pix to make 0-relative array coordinates
        pix = self.wcs.wcs.wcs_world2pix(x, 0)[0]

        if nearest:
            pix = (pix + 0.5).astype(int)
            np.maximum(pix, 0, out = pix)
            if self.shape is not None:
                np.minimum(pix, self.wcs.shape-1, out = pix)
        return pix[0] if np.isscalar(val) else pix

    def pix2wcs(self, pix=None, unit = None):

        if pix is None:
            pixarr = np.arange(self.wcs.shape, dtype = float)
        else:
            pixarr = np.atleast_1d(pix)

        res = self.wcs.wcs.wcs_pix2world(pixarr, 0)[0]
        
        if self.unit == u.Unit("m/s"):
            print("in m/s")
#            unit = u.Unit("km/s")
            res = (res * self.unit).to(u.Unit("km/s")).value
        
        if unit is not None:
            res = (res * self.unit).to(unit).value

        return res[0] if np.isscalar(pix) else res


    def _prep_data(self, interp='no'):
        """
        we may mask values in our view of the data, but `lmfit` will operate on
        the unmasked data. This function fills masked values with 0.

        TODO: include interpolation aspect in case we want to fill them with
        something more natural to the data

        """
        data = ma.filled(self.data, 0.)
        return data

    def _prep_model(self, model_list, unit = True):
        """
        Allows for a model or list of models to be sent to the model prepper

            'g' : Gaussian Model
            'l' : Lorentzian Model
            'v' : Voigt Model

        These allow quick entry of keywords for model types

        TODO: add option for unit conversion, but for now just make `unit` bool
        """
        if unit:
            xunit = self.unit
            yunit = self.flux
        else:
            xunit = u.dimensionless_unscaled
            yunit = u.dimensionless_unscaled

        pix = np.arange(self.wcs.shape, dtype=np.float64)

        xarr = self.pix2wcs(pix) if unit else pix
        res = self._prep_data()

        # Set the model keys
        model_keys = {
                'g' : 'GaussianModel',
                'l' : 'LorentzianModel',
                'v' : 'VoigtModel'
                }

        # set the model_data dict with the data we want to model
        model_data = {
                'x' : xarr,
                'y' : res,
                'unit' : xunit,#spec_unit,
                'model' : []
                }
        # Next, make list where 'type' is key and model_key vals are values
        mlist = []
        for m in model_list:
            m = m.lower()
            mdict = {}
            mdict['type'] = model_keys[m]
            mlist.append(mdict)
        # update the model with keyword data
        model_data.update(model=mlist)

        # make model_data an attribute?
        self.model_info = model_data
        return self.model_info

    def make_model(self, model_list, coords = None, prepped_model = None,
                  unit = True):
        """
        Generate a model using a model list and coord data. This function uses
        `lmfit` to generate the models, perform a least-squares fit, and evaluate
        the fit for plotting and analysis

        Parameters
        -----------
        model_list : list
            list of keywords or model names to pass to the fitter; the number of
            model keywords should equal the length of the `coords` array
        coords : array_like
            list of (x, y) coordinates containing the positions of the initial
            peak guesses
        prepped_model : dict
            a dictionary containing the data, model names, and units; if None,
            a model will be generated
        """

        # set up the model
        if prepped_model is None:
            prep = self._prep_model(model_list)
        else:
            prep = prepped_model.copy()

        composite_model = None
        params = None
        x = prep['x']
        y = prep['y']
        xmin = x.min()
        ymin = y.min()

        # if no coordinates are provided, make a guess.
        # TODO: update this for guessing when there are multiple peaks
        if coords is not None:
            coords = np.asarray(coords)
        elif coords is None and self._coords is not None:
            coords = np.asarray(self._coords)
        else:
            xpix = np.where(self.data == y.max())[0][0]
            xcoord = self.pix2wcs(xpix)
            coords = np.asarray([(xcoord, y.max())])

        # populate the model dictionary with initial parameters
        for i, func in enumerate(prep['model']):
            ctr, peak = coords[i]
            prefix = f'm{i}'
            model = getattr(models, func['type'])(prefix=prefix)

            if func['type'] in ['GaussianModel', 'LorentzianModel',
                                'VoigtModel']:
                model.set_param_hint('amplitude', value = 1.1*peak,
                                      min = 0.5 * peak)
                model.set_param_hint('center', value = ctr, min = ctr-5,
                                      max = ctr + 5)
                model.set_param_hint('sigma', min = 1e-6, max = 10) #max=30
                default_params = {
                        prefix+'center' : ctr,
                        prefix+'height' : peak,
                        prefix+'sigma' : 5
                    }
            else:
                raise NotImplementedError(f"Model {func['type']} not implemented yet")

            # make the parameters
            model_params = model.make_params(**default_params, **func.get('params', {}))

            if params is None:
                params = model_params
            else:
                params.update(model_params)
            if composite_model is None:
                composite_model = model
            else:
                composite_model = composite_model + model

        return composite_model, params

    def fit_model(self, model_list, coords = None, mode = 'components', plot = False,
                  densify = 10, invert_x = False, emline=None, fig_kws=None,
                  ax_kws=None): #unit=True):

        """
        Fit a model or composite model based on user specified parameters.
        This function uses `lmfit` to handle the grunt work.

        Parameters
        -----------
        model_list : list, str
            An ordered list of model keywords used to fill the dict with
            the model types.
        coords : list, tuple
            A list of tuples containing (x, y) coordinates for the peaks
            and their locations
        plot : bool
            Do you want the data plotted?
        densify : int, optional
            optional parameter to make the data dense for plotting fitted curves;
            if not None, the data will be refined by the `densify` factor by:
                dense_array = np.arange(xarr[0], xarr[-1], xarr_step / densify)
            where the `xarr_step` is just the step value of the array

        Returns
        ------------
        out : `lmfit.models.results` or whatever it is
        """
        import matplotlib.pyplot as plt
        # if coords are given, format them and override the class attribute

        if fig_kws is None:
            fig_kws = {'figsize' : (9, 5)}
        if ax_kws is None:
            ax_kws = {'drawstyle' : 'steps-mid', 'linewidth': 1}

        # Get the emission line for the title?
        if emline is not None:
            if emline in lines.keys():
                emis = lines[emline][2]
            else:
                emis = None
        else:
            emis=None

        if coords is not None:
            try:
                coords = np.asarray(coords, dtype = np.float64)
            except ValueError:
                print("Coords must be numeric")
            self._coords = coords
        else:
            try:
                coords = self._coords
            except Exception:
                print("No initial parameters supplied! Estimating from data...")
                #pass

        # make the model
        model_data, params = self.make_model(model_list=model_list, coords=coords)
        xarr = self.model_info['x']
        yarr = self.model_info['y']

        result = model_data.fit(yarr, params, x = xarr)
        self.fit_result = result

        # make a dense array for curve plotting
        x_dense = np.arange(xarr[0], xarr[-1], (xarr[1] - xarr[0])/densify)

        if plot:
            if not plt.get_fignums():
                fig, ax = plt.subplots(**fig_kws)
            else:
                ax = plt.gca()
            #kwargs.update({'drawstyle' : 'steps-mid', 'linewidth': 1})
            if ax.lines:
                pass
            elif ax.collections:
                pass
            else:
                ax.plot(self.model_info['x'], self.model_info['y'], **ax_kws)

            xtype = self.wcs.wcs.wcs.ctype[0]
            xlab = f'{xtype} ({self.unit.to_string("latex")})'
            if invert_x:
                text_offset = -0.5
                x_start = self.wcs.get_stop()
                x_stop = self.wcs.get_start()
                arm = str(' (Blue)')
            else:
                text_offset = 0.5
                x_start = self.wcs.get_start()
                x_stop = self.wcs.get_stop()
                arm = str(' (Red)')
            # Handle motion events?
            def on_move(event):
                if event.inaxes is not None:
                    xc, yc = event.xdata, event.ydata
                    try:
                        i = self.wcs2pix(xc, nearest = True)
                        x = self.pix2wcs(i)
                        event.canvas.toolbar.set_message(
                            'xc=%g yc=%g i=%d dist=%g data=%g' %
                            (xc, yc, i, x, self._data[i]))
                    except Exception as e:
                        print(e) # for debug
                        pass


            # NOTE: all `model_data` instances changed to `self.model_info`
            if mode.lower() == 'components':
                components = result.eval_components(x = x_dense)
                for i, model in enumerate(self.model_info['model']):
                    ax.plot(x_dense, components[f'm{i}'], **ax_kws)
                ax.set_xlabel(xlab)
                ax.set_ylabel(r'Flux ($F_{\lambda}$)')

                if emis:
                    ax.set_title(emis + arm)

                # make centroid labels?
                for key, val in result.params.items():
                    if key.endswith('center'):
                        lab = str(np.round(val, 2)) + self.unit.to_string('latex')
                        ax.axvline(val, ls = ':')
                        plt.text(val + text_offset, y = 0.8 * self.model_info['y'].max(),
                                 s=r' %s'%lab, rotation=90)
                plt.connect('motion_notify_event', on_move)
                #ax.invert_xaxis(invert_x)
                ax.set_xlim(x_start, x_stop)
                plt.tight_layout()

            if mode.lower() == 'residuals':
                print('Do something')

    def get_info(self, convert=True):

        """
        get the basic information about the fit, ie the centroid and HWHM. this
        doesn't worry about converting e.g. angstrom to km/s beccause the main
        PVSlice class will handle this in the `radial_velocity` function.
        """
        from astropy.constants import c

        center = []
        fwhm = []
        sigma = []

        # get the mimimum possible wavelength step of instrument
        minstep = self.wcs.get_step()
        for key, val in self.fit_result.params.items():
            if key.endswith('center'):
                center.append(val.value)
            if key.endswith('fwhm'):
                fwhm.append(val.value)
            if key.endswith('sigma'):
                sigma.append(val.value)

        return np.asarray((center, fwhm, sigma))

    def info(self):
        """
        Print the info if you want it
        """

        log = self._logger.info
        shape_str = (' x '.join(str(x) for x in self.shape)
                    if self.shape is not None else 'no shape')

        data = ('no data' if self.data is None else f'.data({shape_str})')
        xunit = str(self.unit)
        yunit = ('no unit' if self.flux is None else str(self.flux))

        log('%s (%s, %s)', data, xunit, yunit)
        self.wcs.info()

        # has a fit been made?
        if hasattr(self, "fit_result"):
            center, hwhm = np.round(self.get_info(), 2)
            log("Fit Info (in %s)" % self.unit)
            log("%8s %8s" % ('Centroid', 'HWHM'))
            for c, h in zip(center, hwhm):
                log("%8s %8s" % (c, h))



from astropy.io import fits
import astropy.units as u
from numpy import ma
import numpy as np
from lmfit import models
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator

from .data import DataND
from .world import Position, VelWave
from .plotter import *
from .tools import is_notebook
from .lines import lines

class SpecLine(DataND):

    _is_spectral = True

    def mask_range(self, lmin=None, lmax=None, inside=True, unit=None):
        """
        Mask a region inside/outside a range [lmin, lmax]

        Note that this alters the view of the original array; copy the
        object if you want to act on a new one.

        Parameters
        -----------
        lmin : float
            Lower bound; if None, it starts at first pixel
        lmax : float
            Upper bound; if None, it picks the last pixel
        inside : bool
            toggle whether the region within or without the bounds is masked
        unit : `astropy.units.Unit`
            wave/velocity units. None by default, so pass a valid unit or
            the function will go by pixel values
        """

        if self.velwave is None:
            raise ValueError("Can't do this without spectral coordinates")

        else:
            if lmin is None:
                pmin = 0
            else:
                if unit is None:
                    pmin = max(0, int(lmin + 0.5))
                else:
                    pmin = max(0, self.velwave.wav2pix(lmin, nearest = True))
            if lmax is None:
                pmax = self.shape[0]
            else:
                if unit is None:
                    pmax = min(self.shape[0], int(lmax + 0.5))
                else:
                    pmax = min(self.shape[0], self.velwave.wav2pix(lmax, nearest = True))

            if inside:
                self.data[pmin:pmax] = ma.masked
            else:
                self.data[:pmin] = ma.masked
                self.data[pmax+1:] = ma.masked

    def integrate(self, lmin = None, lmax = None, unit = None):
        """
        This will integrate the (non-corrected) flux over a spectral range using
        a simple Simpson's Rule.

        I need to work out the units, but for now this will just return a
        result derived from the native units of the data. Note however that
        the flux over a range will be something like erg/cm^2/s and
        there are no corrections for ie reddening, extinction, etc.

        As of now, the data contains no flux units so the result will be
        simply `u.dimensionless_unscaled * u.unit`. Alternatively, if the units
        of flux are known, simply send them as a keyword when instanciating the
        object or assign the attribute manually, ie `obj.flux_unit = unit`

        Parameters
        ----------
        lmin : float
            lower bound of the velocity or wavelength; if None, the function
            begins at first pixel
        lmax : float
            upper bound; if None, this is the last pixel
        unit : `astropy.units.Unit` or None
            the spectral units of [lmin, lmax]; None by default if these are
            pixel indices

        Returns
        -----------
        out : we'll see
        """

        # get the indices of the specified ranges and all that
        # i1 is index; l1 is wave/velocity
        if lmin is None:
            i1 = 0
            lmin = self.velwave.pix2wav(-0.5)
            #print(f'lmin = {lmin}')
        else:
            if unit is None:
                l1 = lmin
                lmin = self.velwave.pix2wav(max(-0.5, l1))
            else:
                l1 = self.velwave.wav2pix(lmin, nearest=False)
            i1 = max(0, int(l1))

        if lmax is None:
            i2 = self.shape[0]
            lmax = self.velwave.pix2wav(i2 - 0.5)
        else:
            if unit is None:
                i2 = lmax
                lmax = self.velwave.pix2wav(min(self.shape[0] - 0.5, l2))
            else:
                l2 = self.velwave.wav2pix(lmax, nearest=False)
            i2 = min(self.shape[0], int(l2) + 1)

        # to work around the array limits, we'll take the lower wavelength or
        # velocity of each pixel + 1 pixel at the end
        d = self.velwave.pix2wav(-0.5 + np.arange(i1, i2+1))

        # truncate or extend the first and last pixels to the start/end of
        # the values in the spectrum
        d[0] = lmin
        d[-1] = lmax

        if unit is None:
            unit = self.velwave.unit

        # get the data over the range
        data = self.data[i1:i2]

        # do the units agree?
        if unit in self.flux_unit.bases:
            out_unit = self.flux_unit * unit
        else:
            try:
                # sort out flux density
                wunit = (set(self.flux_unit.bases) &
                         set([u.pm, u.angstrom, u.nm, u.um])).pop()

                # scale the wavelength axis
                d *= unit.to(wunit)

                # final units
                out_unit = self.flux_unit * wunit

            # if there's an error anywhere, just return unchanged units
            except Exception:
                out_unit = self.flux_unit * unit

        # standard integration: each pixel value is multiplied by the difference
        # in wavelength from the start of the pixel to the start of the next
        flux = (data * np.diff(d)).sum() * out_unit

        return flux


    def plot(self, **kwargs):
        """
        Placeholder for the moment: sort out plot interactivity etc

        NOTE: calling this function will clear the `_coords` attribute, so be
        careful there.
        """
        if is_notebook():
            self._logger.info('Click on data points to save coordinates.')
        else:
            self._logger.info("To enable point selection on the plot, press "
                          #"'ctrl+p'. Press 'ctrl+q' to disable.")
                              "'ctrl+p'. Press 'ctrl+q' to disable.")

        unit_fmt = u.Unit(self.velwave.unit).to_string('latex')

        coords = []
        # Set up some basic plot interactivity / messages and allow
        # the user to press a key to enter clickable plot mode and
        # save the clicked coordinates to a list
        def on_move(event):
            if event.inaxes is not None:
                xc, yc = event.xdata, event.ydata
                try:
                    #i = self.world.pix2val(xc)
                    i = self.velwave.wav2pix(xc, nearest = True)
                    x = self.velwave.pix2wav(i)
                    #event.canvas.toolbar.set_message(
                    event.canvas.toolbar.set_message(
                        # rf'xc = {xc:0.2f} yc = {yc:0.2f} = {x:0.1f} k = {i} '
                        # rf'data = {self._data[i]:0.2f}' )
                        'xc=%g yc=%g i=%d lbda/vel=%g data=%g' %
                        (xc, yc, i, x, self._data[i]))
                except Exception as e:
                    print(e)
                    pass

        def on_click(event):
            #global ix, iy
            xc, yc = event.xdata, event.ydata
            # ix, iy = event.xdata, event.ydata
            try:
                i = self.velwave.wav2pix(xc, nearest=True)
                x = self.velwave.pix2wav(i)
            #print(f'x={ix}, y={iy}')
                data = self._data[i]
                print(f'v={x}, flux={data}')
                coords.append((x, data))
            #coords.append((ix, iy))
            except Exception as e:
                print(e)
                pass

        def on_key(event):
            if event.key == 'a':#'ctrl+p':
                self._logger.info('Enabling point-selection mode...')
                cid = fig.canvas.mpl_connect('button_press_event', on_click)
            if event.key == 'q': #'ctrl+q':
                self._logger.info("Point-selection mode disabled.")
                cid = fig.canvas.mpl_connect('button_press_event', on_click)
                fig.canvas.mpl_disconnect(cid)
                #plt.disconnect(cid)

        fig, ax = plt.subplots(figsize = (9, 5))
        if self.velwave.unit == u.Unit("m/s"):
            wunit = u.Unit("km/s")
        else: 
            wunit = None
        xarr = self.velwave.pix2wav(unit = wunit)
        data = self.data.copy()
        kwargs.update({'drawstyle' : 'steps-mid', 'linewidth': 1})

        ax.plot(xarr, data, **kwargs)
        plt.connect('motion_notify_event', on_move)
        if is_notebook():
            cid = fig.canvas.mpl_connect('button_press_event', on_click)
        else:
            cid = fig.canvas.mpl_connect('key_press_event', on_key)

        self._coords = coords



from .data import *
from .world import *
from .tools import *
from .pvslice import *
from .spatial import *
from .spectral import *
from .logging import get_logger
from .lines import *

'''
A potentially simple module for handling long-slit and pseudo-long-slit data
'''

#if __name__ == '__main__':#
#    pass
#else:
#    pass

get_logger()



import logging


def get_logger(name='telassar', level='DEBUG',
               msg_fmt='[%(levelname)s] %(message)s', datefmt=None):

    logger = logging.getLogger(name)
    logger.setLevel(level)

    logger.handlers.clear()

    stream = logging.StreamHandler(stream=None)
    stream.setLevel(level)
    formatter = logging.Formatter(msg_fmt, datefmt=datefmt)
    stream.setFormatter(formatter)
    logger.addHandler(stream)



import astropy.units as u
from astropy.io import fits
import numpy as np
from numpy import ma

import logging

from .world import Position, VelWave

class DataND:

    # maybe this will help with sorting the 1D conditions?
    _is_spectral = False
    _is_spatial = False

    def __init__(self, filename=None, data=None, mask=False, dtype=None,
                 ext=None, header=None, unit=None, wcs=None, spec=None,
                 **kwargs):

        self._logger = logging.getLogger(__name__)

        self.filename = filename
        self.ext = ext
        self._data = data
        self._dtype = dtype
        self.header = header or None
        self.position = None
        self.velwave = None
        self.skylines= None

        if (filename is not None) and (data is None):
            # read in a fits file
            hdul = fits.open(filename)

            if len(hdul) == 1:
                self.ext = 0
            elif isinstance(ext, int):
                self.ext = ext

            hdr = hdul[self.ext].header
            self.header = hdr
            self._data = hdul[self.ext].data

            if 'BUNIT' in self.header:
                self.flux_unit = u.Unit(self.header['BUNIT'])
            else:
                self.flux_unit = u.dimensionless_unscaled

            self._mask = ~(np.isfinite(self._data))
            self.position = Position(self.header)
            self.velwave = VelWave(self.header)

        else:
            if mask is ma.nomask:
                self._mask = mask

            if data is not None:
                if self._dtype is None:
                    self._dtype = np.float64

                if isinstance(data, ma.MaskedArray):
                    self._data = np.array(data.data, dtype=np.float64,
                                          copy=True)
                    if data.mask is ma.nomask:
                        self._mask = data.mask
                    else:
                        self._mask = np.array(data.mask, dtype=bool,
                                              copy=True)
                else:
                    self._data = np.array(data, dtype=np.float64,
                                          copy=True)
                    if mask is None or mask is False:
                        self._mask = ~(np.isfinite(data))
                    elif mask is True:
                        self._mask = np.ones(shape=data.shape, dtype=bool)
                    elif mask is not ma.nomask:
                        self._mask = np.array(mask, dtype=bool, copy=True)

            self.flux_unit = u.Unit(u.dimensionless_unscaled)

        # this didn't work like I thought, hopefully there are no issues
        self.set_coords(wcs=wcs, spec=spec)

    @property
    def data(self):
        res = ma.MaskedArray(self._data, mask=self._mask, copy=False)
        return res

    @data.setter
    def data(self, val):
        if isinstance(val, ma.MaskedArray):
            self._data = val.data
            self._mask = val.mask
        else:
            self._data = val
            self._mask = ~(np.isfinite(val))

    @property
    def shape(self):
        return self._data.shape

    @property
    def ndim(self):
        if self._data is not None:
            return self._data.ndim
        try:
            return self.header['NAXIS']
        except KeyError:
            return None

    @property
    def mask(self):
        return self._mask

    @mask.setter
    def mask(self, val):
        if val is ma.nomask:
            self._mask = val
        else:
            self._mask = np.asarray(val, dtype=bool)

    @classmethod
    def new_object(cls, object, data=None, unit=None, wcs=None, spec=None):
        '''
        Copy attributes from one object into a new instance
        Needs testing...

        Parameters
        ----------
        object : `SData`
            Template object
        data : array-like
            If you don't want to copy `object.data`, send a separate array

        '''
        data = object.data if data is None else data
        if unit is None:
            try:
                unit = object.unit
            except AttributeError:
                unit = None

        kwargs = dict(filename=object.filename, data=data, unit=unit,
                      ext=object.ext, header=object.header.copy())

        try:
            kwargs['wcs'] = object.position
            kwargs['spec'] = object.velwave
        except AttributeError:
            kwargs['wcs'] = Position(object.header)
            kwargs['spec'] = VelWave(object.header)
        return cls(**kwargs)

    def copy(self):
        return self.__class__.new_object(self)

    def __repr__(self):
        '''
        For pretty printing
        '''
        if len(self.shape) == 2:
            fmt = """<{}(shape={}, spatial unit = '{}', spectral unit = '{}',
                     dtype = '{}')>"""
            return fmt.format(self.__class__.__name__, self.shape,
                              str(self.position.unit),
                              str(self.velwave.unit).replace(' ', ''),
                              self._dtype)

        elif len(self.shape) == 1 and self._is_spatial:
            fmt = """<{}(shape={}, spatial unit = '{}', dtype = '{}')>"""
            return fmt.format(self.__class__.__name__, self.shape,
                              str(self.position.unit), self._dtype)
        elif len(self.shape) == 1 and self._is_spectral:
            fmt = """<{}(shape={}, spectral unit = '{}', dtype = '{}')>"""
            return fmt.format(self.__class__.__name__, self.shape,
                              str(self.velwave.unit).replace(' ', ''),
                              self._dtype)

    def info(self):

        log = self._logger.info
        shape_str = (' x '.join(str(x) for x in self.shape)
                    if self.shape is not None else 'no shape')
        log('%s (%s)', self.__class__.__name__,
            self.filename or 'no name')

        data = ('no data' if self._data is None else f'Data shape: ({shape_str})')
        spat_unit = ('no unit' if self.position is None else str(self.position.unit))
        spec_unit = ('no unit' if self.velwave is None else str(self.velwave.unit) )

        log('%s (%s, %s)', data, spat_unit, spec_unit.replace(' ', ''))

        if self.position is None:
            log('No world coordinates installed')
        else:
            self.position.info()

        if self.velwave is None:
            log('No spectral coordinates installed')
        else:
            self.velwave.info()

    def __getitem__(self, item):

        '''
        Return a sliced object. 
        NOTE: This doesn't work the way I want
        '''
        data = self._data[item]
        mask = self._mask
        if mask is not ma.nomask:
            mask = mask[item]
        filename = self.filename
        reshape = None

        if self.ndim == 2:
            # handle a PVSlice[ii,jj] where ii, jj can be int or slice
            # objects
            if isinstance(item, (list, tuple)) and len(item) == 2:
                try:
                    wcs = self.position[item[0]]
                    spec = self.velwave[item[1]]
                    #print(wcs)
                except Exception:
                    print('No WCS information available')
                    wcs = None
                    spec = None
                #print(wcs)
                if isinstance(item[0], int) != isinstance(item[1], int):
                    if isinstance(item[0], int):
                        reshape = (1, data.shape[0])
                    else:
                        reshape = (data.shape[0], 1)

            elif isinstance(item, (int, slice)):

                try:
                    wcs = self.position[item[0], slice(None)]
                    spec = self.velwave[item[1], slice(None)]
                except Exception:
                    wcs = None
                    spec = None

                if isinstance(item, int):
                    reshape = (1, data.shape[0])

            elif item is not None or item == ():
                try:
                    wcs = self.position.copy()
                    spec = self.velwave.copy()
                except Exception:
                    wcs = None
                    spec = None

        elif self.ndim == 1:
            # handle a spatial or spectral profile
            if isinstance(item, slice):
                try:
                    wcs = self.position[item]
                except Exception:
                    wcs = None
                try:
                    spec = self.velwave[item]
                except Exception:
                    spec = None
            elif item is None or item == ():
                try:
                    wcs = self.position.copy()
                except Exception:
                    wcs = None
                try:
                    spec = self.velwave.copy()
                except Exception:
                    spec = None
        # do we need to reshape?
        if reshape is not None:
            data = data.reshape(reshape)
            if mask is not ma.nomask:
                mask = mask.reshape(reshape)

        return self.__class__(
            filename=filename, data=data, mask=mask, dtype=self._dtype,
            ext=self.ext, header=self.header, wcs=wcs, spec=spec)

    def min(self):
        '''
        return minimum unmasked value in the data
        '''
        res = ma.min(self.data)
        return res

    def max(self):
        '''
        return maximum unmasked value in data
        '''
        res = ma.max(self.data)
        return res

    def set_coords(self, wcs=None, spec=None):
        """
        Set the wcs info for the object. Hopefully this sorts the issue
        of reducing 2D PV data to 1D spatial/spectral data?
        """

        if self.header is not None:
            hdr = self.header.copy()
        else:
            hdr = None

        # Install PV coordinates. This checks first if `wcs` or `velwave` args
        # are passed and installs them; if either is`None`, then it checks for
        # a header and attempts to install coords from that; otherwise, raise
        # an error
        if len(self.shape) == 2:
            try:
                if (wcs is not None) and (spec is not None):
                    self.position = wcs
                    self.velwave = spec
                elif (wcs is None or spec is None) and (hdr is not None):
                    self.position = Position(hdr)
                    self.velwave = VelWave(hdr)
            except Exception:
                    self._logger.warning("Unable to install coordinates",
                                        exc_info=True)
                    self.position = None
                    self.velwave = None

        # If the data is 1D, sort out which is which
        if len(self.shape) ==1:
            if self._is_spatial:
                try:
                    if wcs is not None:
                        self.position = wcs
                    elif wcs is None and hdr is not None:
                        self.position = Position(hdr)
                except Exception:
                    self._logger.warning("Unable to install spatial "
                                        "coordinates", exc_info=True)
                    self.position = None
            if self._is_spectral:
                try:
                    if spec is not None:
                        self.velwave = spec
                    elif spec is None and hdr is not None:
                        self.velwave = VelWave(hdr)
                except Exception:
                    self._logger.warning("Unable to install spatial "
                                        "coordinates", exc_info=True)
                    self.position = None

    def unmask(self):
        '''Unmask all but invalid data'''
        if self.mask is not ma.nomask:
            self._mask = ~np.isfinite(self._data)



import astropy.units as u
import numpy as np
from numpy import ma
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator, MaxNLocator
from .tools import timeit


class ImCoords:

    def __init__(self, image, data, toggle_unit=False):
        self.image = image
        self.data = data
        self.toggle_unit = toggle_unit


    def __call__(self, x, y):

        im = self.image

        # figure out if the image passed pixel units or data units
        if self.toggle_unit:
            # get the pixel values
            col = im.velwave.wav2pix(x, nearest=True)
            row = im.position.offset2pix(y, nearest= True)

        else:
            col = int(x + 0.5)
            row = int(y + 0.5)

        if (im.position is not None and im.velwave is not None and row >=0
                and row < im.shape[0] and col >= 0 and col < im.shape[1]):
            #print(f'{val} is scalar')

            xc = im.velwave.pix2wav(col, unit = im.velwave.unit)
            yc = im.position.pix2offset(row, unit = im.position.unit)
            val = self.data[row, col]

            if np.isscalar(val):
                return 'y=%g x=%g p=%i q=%i data=%g' % (yc, xc, row, col, val)
            else:
                return 'y=%g x=%g p=%i q=%i data=%s' % (yc, xc, row, col, val)
        else:
            return 'x=%1.4f, y=%1.4f' % (x, y)

def get_plot_norm(data, vmin = None, vmax = None, zscale = False, scale = 'linear'):
    from astropy import visualization as viz
    from astropy.visualization.mpl_normalize import ImageNormalize

    if zscale:
        interval = viz.ZScaleInterval()
        try:
            vmin, vmax = interval.get_limits(data.filled(0))
        except AttributeError:
            vmin, vmax = interval.get_limits(data)


    if scale == 'linear':
        stretch = viz.LinearStretch
    elif scale == 'log':
        stretch = viz.LogStretch
    elif scale in ('asinh', 'arcsinh'):
        stretch = viz.AsinhStretch
    elif scale == 'sqrt':
        stretch = viz.SqrtStretch
    else:
        raise ValueError('Unknown scale: {}'.format(scale))

    norm = ImageNormalize(vmin = vmin, vmax = vmax, stretch = stretch(), clip = False)

    return norm

def get_plot_extent(wcs_obj, spec_obj):

    '''
    Assuming a `PVSlice.world` object is passed, get the extents for plotting
    '''
    xmin = spec_obj.get_start()
    xmax = spec_obj.get_stop()
    ymin = wcs_obj.get_start()
    ymax = wcs_obj.get_stop()
    
    if spec_obj.wcs.wcs.ctype[0] == 'VELO':
        xmin /= 1000
        xmax /= 1000
    return xmin, xmax, ymin, ymax

def get_background_rms(data, sigma=3, N=10, mask = None):
    '''
    Get the background rms/sigma value of the data. We consider a
    3sigma value as the detection limit, so it is useful to scale
    contour plots using this value to emphasize source detections.

    Parameters
    ----------
    data : `np.ndarray` or `np.ma.MaskedArray`
        the data to be plotted
    sigma : int
        the sigma value sent to `SigmaClip`. data above or below sigma stddevs
        will be ignored in computing the background RMS; default is 3.
    N : int
        the number of sampling boxes to fit in the data. Default is 10. if the
        box sizes are a non-integer value, then the `edge_method` keywork used
        in `Background2D` is set to "pad"
    mask : `np.ma.MaskedArray` or None
        if no mask is specified,

    '''

    from astropy.stats import SigmaClip
    from photutils import Background2D, MedianBackground, StdBackgroundRMS
    
    # get a boxsize. sides must be integer values, so if there's a
    # remainder then set an edge_method keyword
    ny, nx = data.shape
    sy = ny // N
    sx = nx // N

    if (ny % N != 0) or (nx % N != 0):
        edge_method = 'pad'
    else:
        edge_method = None

    # is there a mask?
    if mask is None:
        try:
            mask = data.mask
        except Exception:
            mask = ~(np.isfinite(data))
    else:
        mask = mask
    #mask = data.mask if mask is None else mask

    # do we need to unmask the data?
    # TODO: why does data[~data.mask] compress data to 1D?
    # reshape if necessary I guess
    try:
        ndata = data[~data.mask]
        #print(ndata.shape)
        if ndata.shape != data.shape:
            ndata = ndata.reshape(data.shape)
        #print(ndata.shape)
    except Exception as e:
        print('Exception: ', e)
        ndata = ma.getdata(data)

    #print(ndata.shape)
    # get the background RMS. This calls `photutils.Background2D`
    # for now it doesn't work the way I thought because I don't
    # know what I'm doing, but whatever
    sigma_clip = SigmaClip(sigma=sigma)
    bkg_estimator = MedianBackground()
    rms_estimator = StdBackgroundRMS()
    bkg = Background2D(ndata, (sy, sx), filter_size=(5,5), sigma_clip=sigma_clip,
                    bkg_estimator=bkg_estimator, bkgrms_estimator=rms_estimator,
                    mask=mask, edge_method=edge_method)

    sigrms = bkg.background_rms_median

#    sigrms = sigma_clip(ndata, axis=0).mean()
    return sigrms

def get_contour_levels(data, sigma):
    '''
    Generate contour levels?
    '''
    # quick upper/lower bound for exponents
    # Contours typically start between 3sigma and 5sigma and are based on a
    # sqrt(2) log scale, such that
    #   lvls = (4/3) * sigma * sqrt(2)^x
    # which begins the levels at 4sigma
    # By default, the lower and upper levels are 0.017 and 1, and the exponents
    # are found by:
    #   x = (2 * ln(3 * lvls / sigma) / ln(2) - 4
    expo = lambda lvl, sig : (2 * np.log(3 * lvl/sig) / np.log(2)) - 4

    ndata = data.copy()
    #ndata[ndata < -6*sigma] = ma.masked
    # get min and max of data
    dmax = ndata.max()
    dmin = ndata.min()
    # get upper and lower contour levels
    l0 = 0.01697583 * dmax
    l1 = 1. * dmax
    
    # get upper and lower exponent bounds
    n0 = expo(l0, sigma)
    n1 = expo(l1, sigma)
    # get range of exponents. by default, ,there are 7 levels
    xp = np.linspace(n0, n1, 7)

    # get primary and background contour levels
    lvls1 = (4/3) * (sigma) * np.sqrt(2)**xp
    lvls2 = np.linspace(dmin, 0.8 * sigma, 9)

    #scale = np.array([0.01697583, 0.03395166, 0.06790333, 0.13580666,
    #                  0.27161332, 0.54322664, 1.08645327])
    #lvls1 = dmax * scale

    return lvls1, lvls2

def configure_axes(ax, obj):

    labels = {
        'OFFSET' : 'Offset',
        'VELO' : 'V$_{rad}$',
        'WAVE' : '$\lambda$',
        'AWAV' : '$\lambda$'
    }

    spec = obj.velwave
    pos = obj.position

    spec_type = spec.wcs.wcs.ctype[0]
    spat_type = pos.wcs.wcs.ctype[0]

    y_type = labels[spat_type] if spat_type in labels.keys() else ''
    x_type = labels[spec_type] if spec_type in labels.keys() else ''

    ax.set_xlabel(rf'{x_type} ({u.Unit(spec.unit).to_string("latex")})', fontsize = 10)
    ax.set_ylabel(rf'{y_type} ({u.Unit(pos.unit).to_string("latex")})', fontsize = 10)

    # because the formatting can be weird
    xsize = spec.pix2wav().size

    ax.margins(0.05)
    ax.xaxis.set_major_locator(MaxNLocator(6))
    ax.xaxis.set_minor_locator(AutoMinorLocator())
    ax.yaxis.set_minor_locator(AutoMinorLocator())
    ax.tick_params(which = 'major', direction = 'in', length = 9)
    ax.tick_params(which = 'minor', direction = 'in', length = 6)



import numpy as np
import astropy.units as u
import csv
from functools import wraps
from time import time
import collections

from astropy.wcs import WCS, WCSSUB_SPECTRAL

"""
This is just an assortment of possibly useful tools that can be called
from anywhere. Mostly to keep clean.
"""

def timeit(f, *args, **kwargs):
    @wraps(f)
    def timed(*args, **kwds):
        t0 = time()
        res = f(*args, **kwds)
        t1 = time()
        return res
    return timed

def solve_linear_set(matrix, vector):
    '''
    Like Cramer's rule but easier, and you get the solution
    immediately instead of handling determinants first
    '''
    matrix = np.array(matrix)
    inv_mat = np.linalg.inv(matrix)
    B = np.array(vector)
    x = np.dot(inv_mat, B)
    return x

def is_notebook():
    """
    Are we in a jupyter notebook?
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True
        elif shell == 'TerminalInteractiveShell':
            return False
        else:
            return False
    except NameError:
        return False

def progress_bar(*args, **kwargs):
    from tqdm import tqdm, tqdm_notebook
    func = tqdm_notebook if is_notebook() else tqdm
    return func(*args, **kwargs)

def get_simple_slice(center, slit_width, dshape, slit_length = None):

    '''
    Return a slice or rectangle or whatever based on the length and width
    you want

    Parameters
    ----------
    center : tuple
        the center in (y, x) format from which to make the cut
    slit_width : int, float
        the width of the cut in pixels
    dshape : tuple
        make sure the cut doesn't extend beyond the extent of the data
    slit_length : int or float
        the length of the slit; if None, the length is determined from the
        shape
    '''

    # convert width to two radii
    ny, nx = dshape

    # make sure the center is an array
    center = np.asarray(center)

    # the radius in x is half the slit width
    rx = width / 2.

    # make sure the width is in the limits of the cube
    xmin = max(0, np.floor((center[1] - rx) + 0.5).astype(int))
    xmax = min(nx, np.floor((center[1] + rx) + 0.5).astype(int))

    # if a length is specified, find where its lower and upper
    # bounds are in the data.
    # For a reference point (xc, yc) and slit of length L orientated
    # along the y-axis, the slit extends L/2 in either direction
    # from yc. The upper and lower bounds of the slit are a distance
    # (yc +/- L/2) from the origin
    if length is not None:
        lw = length / 2.
        ymin = max(0, np.floor((center[0] - lw) + 0.5).astype(int))
        ymax = min(ny, np.floor((center[0] + lw) + 0.5).astype(int))
    else:
        ymin = 0
        ymax = ny

    simple_slice = [ slice(ymin, ymax+1), slice(xmin, xmax+1) ]

    return simple_slice

def format_header(hdr, mode):

    hdu = hdr.copy()
    #print(hdu)
    # get some values and make some assumptions
    # `cunit` can be discovered from the mode
    # `cdelt` is, for now, either 0.2 (for arcsec) or 1.25 (for angstrom)
    # because this is how MUSE handles it. Setters exist to change these
    # values. If no mode is passed, `cdelt` is 1. and `cunit` is 'pixel'
    cunit = get_unit_from_mode(mode)
    #print('cunit = ', cunit)
    ctype = "LINEAR"

    if cunit is u.arcsec:
        cdelt = 0.2
    elif cunit is u.angstrom:
        cdelt = 1.25
        ctype = 'AWAV'
    else:
        cunit = u.Unit('pixel')
        cdelt = 1.

    # make essential keywords in case the header is minimal
    # we're assuming `crpix` and `crval` to be at the origin for simplicity
    hdr_keys = {
        'CRPIX' : 1.,
        'CRVAL' : 0.,
        'CDELT' : cdelt,
        'CUNIT' : cunit.to_string('fits'),
        'CTYPE' : ctype
    }

    # if the important keywords are not in the header, add them
    # again this is for a simple 1D case! we aren't handling n > 1.
    n = hdu['NAXIS']

    for key, val in hdr_keys.items():
        if f'{key}{n}' not in hdu:
            hdu.set(f'{key}{n}', val)

    return hdu

def read_modlist_from_csv(fname):
    '''
    If your model list is in a CSV file, here's a way to pick it out

    Parameters
    -----------
    fname : str
        Obviously a filename; may add more kwargs later for an option to use
        `pandas`

    Returns
    -----------
    out : `OrderedDict`
        A dictionary of the CSV, with headers being the keys and the columns
        being the entries
    '''
    modlist = collections.defaultdict(list)

    with open(fname) as f:
        reader = csv.DictReader(f, delimiter=',')
        for row in reader:
            # Are some columns longer than others? Skip the blank cells
            if any(row[key] in (None, '') for key in row):
                continue
            # Only need the first letter since that's how we roll
            for key in row:
                modlist[key].append(row[key][0])

    return modlist


def parse_badlines(fname):
    '''
    A convenience function to parse and install skyline information
    from a data file
    '''
    with open(fname) as f:
        for line in f.readlines():
            l1, l2, emis = line.strip().split()
            yield emis, float(l1), float(l2)



"""
Modified from:
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')


setup(
    name='telassar',  # Required
    version='0.0.1',  # Required
    description='Two dimEnsionaL spectrAl analysiS for muSe And xshooteR',  # Optional
    long_description=long_description,  # Optional
    long_description_content_type='text/markdown',  # Optional (see note above)
    url='https://github.com/amiller361/telassar',  # Optional
    author='Andrew Miller',  # Optional
    author_email='andrew.miller@mu.ie',  # Optional
    keywords='astronomy, spectroscopy, astrophysics',  # Optional
    package_dir={'': 'telassar'},  # Optional
    packages=find_packages(where='telassar'),  # Required
    python_requires='>=3.6',
    install_requires=[
        'astropy',
        'matplotlib',
        'lmfit',
        'photutils',
    ],  # Optional

    # include sample data?
    data_files=[('hd163296_pvslice.fits', ['data/hd163296_pvslice.fits'])],        
#    entry_points={  # Optional
#        'console_scripts': [
#            'sample=sample:main',
#        ],
#    },

     project_urls={  # Optional
        'Bug Reports': 'https://github.com/amiller361/telassar/issues',
#        'Say Thanks!': 'http://saythanks.io/to/example',
        'Source': 'https://github.com/amiller361/telassar/',
    },
)



from astropy.io import fits
import astropy.units as u
from astropy.wcs import WCS as pywcs
from astropy.wcs import WCSSUB_SPECTRAL
import numpy as np

import logging

def get_cdelt_from_cd(cd):
    '''Get the cdelt from the CD matrix'''


    dy = np.sqrt(cd[0,1]**2 + cd[1,1]**2)
    dx = np.sqrt(cd[1,0]**2 + cd[0,0]**2)
    if len(cd) == 3:
        dz = np.sqrt(cd[2,0]**2 + cd[2,2]**2)

        cdelt = np.array([dz, (dy*u.deg).to(u.arcsec).value])
    else:
        cdelt = np.array([dx, dy])
#    print(cdelt)
    return cdelt


def wcs_from_header(header):

    hdr = header.copy()

    cunit = None
    try:
        for i in range(1, hdr['NAXIS']+1):
            if 'angstrom' in hdr['CUNIT%d' %i].lower():
                cunit = u.Unit(hdr.pop('CUNIT%d'%i))
    except KeyError:
        pass
    # if 'CUNIT3' in hdr:
    #     cunit = u.Unit(hdr.pop('CUNIT3'))
    # elif 'CUNIT2' in hdr:
    #     cunit = u.Unit(hdr.pop('CUNIT2'))
    # else:
    #     cunit = u.Unit('angstrom')

    try:
        n = hdr['NAXIS']
    except KeyError:
        try:
            n = hdr['WCSAXES']
        except KeyError:
            print("Can't install coordinates!")
            return

    # generate WCS object
    mywcs = pywcs(hdr, fix=False)
    old_shape = mywcs.pixel_shape

    if mywcs.wcs.has_cd():
        #print('Has CD')
        #try:
        cdelt = get_cdelt_from_cd(mywcs.wcs.cd)
        #except IndexError:
        #    cdelt = mywcs.wcs.cd
#        print(cdelt)

    elif mywcs.wcs.has_pc():
        #print('Has PC')
        try:
            cd = np.dot(np.diag(mywcs.wcs.get_cdelt()), mywcs.wcs.get_pc())
            cdelt = get_cdelt_from_cd(cd)
        except IndexError:
            cdelt = mywcs.wcs.get_cdelt()

    if mywcs.wcs.has_cd():
        del mywcs.wcs.cd

    if n==3:
        nz, ny = old_shape[::-1][:2]
        new_wcs = mywcs.sub([WCSSUB_SPECTRAL, 0])
        new_wcs.pixel_shape = (nz, ny)

    if n==2:
        new_wcs = mywcs.copy()

    crpix2 = hdr['CRPIX2']

    ctype2 = 'OFFSET'

    # is this velocity or wavelength?
    if new_wcs.wcs.crval[0] < 0.:
        ctype1 = 'VELO'
    else:
        ctype1 = 'AWAV'

    new_wcs.wcs.crpix = [1., crpix2]
    new_wcs.wcs.cdelt = cdelt
    new_wcs.wcs.ctype = [ctype1, ctype2]

    # I don't know why it's necessary, but otherwise the step gets calculated
    # as the cd**2
    new_wcs.wcs.pc = np.diag([1., 1.])


    #new_wcs.wcs.cunit[1] = u.Unit('arcsec')
    return new_wcs

class Position:

    def __init__(self, hdr=None, crpix=1., crval=1., cdelt=0.2,
                 unit=u.arcsec, ctype='LINEAR', shape=None):

        self._logger = logging.getLogger(__name__)

        unit = u.Unit(unit).to_string('fits')

        self.shape = shape

        if (hdr is not None):
            h = hdr.copy()

            # the spatial axis should be axis=2
            axis = 2
            self.wcs = wcs_from_header(h).sub([axis])

            if self.wcs.wcs.cunit[0] != '':
                self.unit = self.wcs.wcs.cunit[0]
            elif self.wcs.wcs.ctype[0] == 'OFFSET':
                self.unit = u.Unit('arcsec')

        elif hdr is None:
            self.unit = u.Unit(unit)
            self.wcs = pywcs(naxis=1)

            # set the reference pixel
            self.wcs.wcs.crval[0] = crval
            self.wcs.wcs.ctype[0] = ctype
            self.wcs.wcs.cdelt[0] = cdelt
            self.wcs.wcs.crpix[0] = crpix
            self.wcs.pixel_shape = (shape,)

        try:
            self.shape = self.wcs.pixel_shape[0] if shape is None else shape
        except Exception:
            self._logger.warning('No shape is provided')

    def copy(self):
        """
        Copy the Position object
        """
        out = Position(shape=self.shape, unit=self.unit)
        out.wcs = self.wcs.deepcopy()
        return out

    def __repr__(self):
        return repr(self.wcs)

    def info(self, unit=None):
        try:
            unit = unit or self.unit
            start = self.get_start(unit=unit)
            step = self.get_step(unit=unit)
            type = self.wcs.wcs.ctype[0].capitalize()

            if self.shape is None:
                self._logger.info('Spatial %s: min: %0.2f" step: %0.3f"' %
                                 (type, start, step))
            else:
                end = self.get_stop(unit=unit)
                self._logger.info('Spatial %s: min: %0.1f" max: %0.1f" step: %0.3f"' %
                                 (type, start, end, step))
        except Exception as e:
            print(e)
            self._logger.info("something happened I can't fix yet")

    def __getitem__(self, item):

        if item is None:
            return self

        elif isinstance(item, int):
            if item >= 0:
                arc = self.pix2offset(pix=item)
            else:
                if self.shape is None:
                    raise ValueError("Can't return an index without a shape")
                else:
                    arc = self.pix2offset(pix=self.shape + item)
            return Position(crpix=1., crval=arc, cdelt=0., unit=self.unit,
                            ctype=self.wcs.wcs.ctype[0], shape=1)

        elif isinstance(item, slice):
            if item.start is None:
                start = 0
            elif item.start >= 0:
                start = item.start
            else:
                if self.shape is None:
                    raise ValueError("Can't return an index without a shape")
                else:
                    start = self.shape + item.start
            if item.stop is None:
                if self.shape is None:
                    raise ValueError("Can't return an index without a shape")
                else:
                    stop = self.shape
            elif item.stop >= 0:
                stop = item.stop
            else:
                if self.shape is None:
                    raise ValueError("Can't return an index without a shape")
                else:
                    stop = self.shape + item.stop
            newarc = self.pix2offset(pix=np.arange(start, stop, item.step))
            dimens = newarc.shape[0]

            if dimens < 2:
                raise ValueError("Offset with dimension < 2")
            cdelt = newarc[1] - newarc[0]
            return Position(crpix=1., crval=newarc[0], cdelt=cdelt,
                            unit=self.unit, ctype=self.wcs.wcs.ctype[0],
                            shape=dimens)
        else:
            raise ValueError("Can't do it!")

    def offset2pix(self, val, nearest=False):
        """
        Return a pixel value if given an offset value in
        arcseconds
        """
        x = np.atleast_1d(val)

        # tell world2pix to make 0-relative array coordinates
        pix = self.wcs.wcs_world2pix(x, 0)[0]

        if nearest:
            pix = (pix + 0.5).astype(int)
            np.maximum(pix, 0, out=pix)
            if self.shape is not None:
                np.minimum(pix, self.shape-1, out=pix)
        return pix[0] if np.isscalar(val) else pix

    def pix2offset(self, pix=None, unit=None):
        """
        Reverse of above: get spatial value of pixel
        """

        if pix is None:
            pixarr = np.arange(self.shape, dtype=float)
        else:
            pixarr = np.atleast_1d(pix)

        res = self.wcs.wcs_pix2world(pixarr, 0)[0]

        if unit is not None:
            res = (res * self.unit).to(unit).value

        return res[0] if np.isscalar(pix) else res


    def get_step(self, unit=None):
        '''
        get the step
        '''
        if self.wcs.wcs.has_cd():
            step = self.wcs.wcs.cd[0]
        else:
            cdelt = self.wcs.wcs.get_cdelt()[0]
            pc = self.wcs.wcs.get_pc()[0][0]
            step = cdelt #* pc

        if unit is not None:
            step = (step * self.unit).to(unit).value
        return step

    def set_step(self, x, unit=None):
        """
        If you want to change the step, do it here;
        useful if you've read in an array and now have minimal header data
        """

        if unit is not None:
            step = (x * unit).to(self.unit).value
        else:
            step = x

        # TODO: probably won't have any CD info for 1D files
        # so maybe just ignore this?
        if self.wcs.wcs.has_cd():
            self.wcs.wcs.cd[0][0] = step
        else:
            pc = self.wcs.wcs.get_pc()[0][0]
            self.wcs.wcs.cdelt[0] = step / pc

    def get_start(self, unit=None):
        """
        Get the starting pixel value
        """
        if unit is not None:
            return (self.pix2offset(0)* unit).to(self.unit).value
        else:
            return self.pix2offset(0)

    def get_stop(self, unit=None):

        if self.shape is None:
            raise IOError("Need a dimension!")
        else:
            if unit is not None:
                return (self.pix2offset(self.shape - 1) * unit).to(self.unit).value
            else:
                return self.pix2offset(self.shape - 1)

    def get_range(self, unit=None):
        '''
        Simply return the upper and lower bounds of the array
        '''
        try:
            return self.pix2offset([0, self.shape-1], unit)
        except AttributeError:
            print("no dimension provided for spatial array!")


class VelWave:

    def __init__(self, hdr=None, crpix=1., crval=1., cdelt=1.,
                 unit=u.angstrom, ctype='LINEAR', shape=None):

        self._logger = logging.getLogger(__name__)

        unit = u.Unit(unit)

        self.shape = shape

        if (hdr is not None):
            h = hdr.copy()

            # check if there's an instrument key in the header
            if 'INSTRUME' in hdr:
                if hdr['INSTRUME'] == 'XSHOOTER':
                    unit = u.Unit(u.nm)
            # the spectral axis should be 1
            axis = 1
            self.wcs = wcs_from_header(h).sub([axis])

            if self.wcs.wcs.ctype[0] == 'VELO':
                self.unit = u.Unit('km/s')
            elif self.wcs.wcs.ctype[0] == 'AWAV' or 'WAVE':
                self.unit = unit
            elif self.wcs.wcs.ctype[0] == 'LINEAR':
                self.unit = unit

        elif hdr is None:
            self.unit = u.Unit(unit)
            self.wcs = pywcs(naxis = 1)

            # set the reference pixel
            self.wcs.wcs.crval[0] = crval
            self.wcs.wcs.ctype[0] = ctype
            self.wcs.wcs.cdelt[0] = cdelt
            self.wcs.wcs.crpix[0] = crpix
            # this is weird but go with it
            self.wcs.pixel_shape = (shape,)

        try:
            self.shape = self.wcs.pixel_shape[0] if shape is None else shape
        except Exception:
            self._logger.warning('No shape is provided')

    def __repr__(self):
        return(repr(self.wcs))

    def info(self, unit=None):

        unit = unit or self.unit
        start = self.get_start(unit=unit)
        step = self.get_step(unit=unit)
        type = self.wcs.wcs.ctype[0].capitalize()

        if self.shape is None:
            unit = str(unit).replace(' ', '')
            self._logger.info('Spectral extent: min: %0.2f %s step: %0.3f %s' %
                             (start, unit, step,  unit))
        else:
            end = self.get_stop(unit=unit)
            unit = str(unit).replace(' ', '')
            self._logger.info('Spectral extent: min %0.2f %s max: %0.2f %s step: %0.3f %s' %
                             (start, unit, end, unit, step, unit))

    def __getitem__(self, item):

        if item is None:
            return self

        elif isinstance(item, int):
            if item >= 0:
                val = self.pix2wav(pix=item)
            else:
                if self.shape is None:
                    raise ValueError("Can't return an index without a shape")
                else:
                    val = self.pix2wav(pix=self.shape + item)
            return Position(crpix=1., crval=val, cdelt=0., unit=self.unit,
                            ctype=self.wcs.wcs.ctype[0], shape=1)

        elif isinstance(item, slice):
            if item.start is None:
                start = 0
            elif item.start >= 0:
                start = item.start
            else:
                if self.shape is None:
                    raise ValueError("Can't return an index without a shape")
                else:
                    start = self.shape + item.start
            if item.stop is None:
                if self.shape is None:
                    raise ValueError("Can't return an index without a shape")
                else:
                    stop = self.shape
            elif item.stop >= 0:
                stop = item.stop
            else:
                if self.shape is None:
                    raise ValueError("Can't return an index without a shape")
                else:
                    stop = self.shape + item.stop
            newval = self.pix2wav(pix=np.arange(start, stop, item.step))
            dimens = newval.shape[0]

            if dimens < 2:
                raise ValueError("Velocity/Wavelength with dimension < 2")
            cdelt = newval[1] - newval[0]
            return VelWave(crpix=1., crval=newval[0], cdelt=cdelt,
                            unit=self.unit, ctype=self.wcs.wcs.ctype[0],
                            shape=dimens)
        else:
            raise ValueError("Can't do it!")

    def wav2pix(self, val, nearest=False):
        """
        Return a pixel value if given an offset value in
        arcseconds
        """
        x = np.atleast_1d(val)

        # tell world2pix to make 0-relative array coordinates
        pix = self.wcs.wcs_world2pix(x, 0)[0]

        if nearest:
            pix = (pix + 0.5).astype(int)
            np.maximum(pix, 0, out=pix)
            if self.shape is not None:
                np.minimum(pix, self.shape-1, out=pix)
        return pix[0] if np.isscalar(val) else pix

    def pix2wav(self, pix=None, unit=None):

        if pix is None:
            pixarr = np.arange(self.shape, dtype=float)
        else:
            pixarr = np.atleast_1d(pix)

        res = self.wcs.wcs_pix2world(pixarr, 0)[0]

        if unit is not None:
            res = (res * self.unit).to(unit).value

        return res[0] if np.isscalar(pix) else res

    def get_start(self, unit=None):
        return self.pix2wav(0)

    def get_stop(self, unit=None):

        if self.shape is None:
            raise IOError("Need a dimension!")
        else:
            return self.pix2wav(self.shape - 1)

    def get_step(self, unit=None):
        '''
        get the step
        '''
        if self.wcs.wcs.has_cd():
            step = self.wcs.wcs.cd[0]
        else:
            cdelt = self.wcs.wcs.get_cdelt()[0]
            pc = self.wcs.wcs.get_pc()[0][0]
            step = cdelt * pc

        if unit is not None:
            step = (step * self.unit).to(unit).value
        return step

    def set_step(self, x, unit=None):
        """
        If you want to change the step, do it here;
        useful if you've read in an array and now have minimal header data
        """

        if unit is not None:
            step = (x * unit).to(self.unit).value
        else:
            step = x

        # TODO: probably won't have any CD info for 1D files
        # so maybe just ignore this?
        if self.wcs.wcs.has_cd():
            self.wcs.wcs.cd[0][0] = step
        else:
            pc = self.wcs.wcs.get_pc()[0][0]
            self.wcs.wcs.cdelt[0] = step / pc

    def get_range(self, unit=None):
        '''
        Simply return the upper and lower bounds of the array
        '''
        try:
            return self.pix2wav([0, self.shape-1], unit)
        except AttributeError:
            print("no dimension provided for spectral array!")
